/*====================================================================

OpenCCM: The Open CORBA Component Model Platform
Copyright (C) 2000-2002 USTL - LIFL - GOAL
Contact: openccm@objectweb.org

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
USA

Initial developer(s): Christophe Contreras
Contributor(s): ___________________________________________________.

====================================================================*/

package org.objectweb.zeusx;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Writer;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Vector;
import java.util.BitSet;

// Zeus imports
import org.enhydra.zeus.Binding;
import org.enhydra.zeus.Generator;
import org.enhydra.zeus.ZeusDefaults;
import org.enhydra.zeus.ZeusException;
import org.enhydra.zeus.binding.AtomicProperty;
import org.enhydra.zeus.binding.BaseBinding;
import org.enhydra.zeus.binding.Container;
import org.enhydra.zeus.binding.Property;
import org.enhydra.zeus.util.CapitalizationUtils;
import org.enhydra.zeus.util.ClassUtils;
import org.enhydra.zeus.util.NamingUtils;
import org.enhydra.zeus.util.XMLName;

import org.enhydra.zeus.generator.*;


/**
 * <p>
 *  <code>{@link Generator}</code> is the portion of Zeus that will convert
 *    a set of constraints, visible through a <code>{@link Binding}</code>,
 *    and output them as generated Java classes (to a
 *    <code>File</code> [directory]).
 * </p><p>
 *  <code>ObjectWebGUIsGenerator</code> is the ObjectWeb XmlTool additional
 *    <code>Generator</code> and takes in either a single <code>Binding</code>
 *    or a <code>List</code> of them, and directly generates Java classes.
 * </p>
 *
 * @author Christophe Contreras
 */
public class ObjectWebGUIsGenerator implements Generator {

    /** A list of bindings to ignore */
    protected List ignoredBindings;

    /** This instances's collapsed elements */
    protected List collapsedElements;

    /** Base directory for generated files. */
    protected File outputDir;

    /** package where we'll find the interfaces. */
    protected String interfacesPackage;

    /** directory where we'll find the models. */
    protected String modelsDirectory;

    protected ModelBindings modelBindings;

    /** Keeps track of whether the Generator needs to be reinitialized. */
    protected boolean initialized;

    /**
     * <p>
     *  This will construct a new instance of <code>ObjectWebGUIsGenerator</code>.
     * </p>
     */
    public ObjectWebGUIsGenerator(String wantedInterfacesPackage, String wantedModelsDirectory, ModelBindings modelBindings) {
        initialized = false;
        ignoredBindings = new LinkedList();
        collapsedElements = new LinkedList();

        this.interfacesPackage = wantedInterfacesPackage;
        this.modelsDirectory = wantedModelsDirectory;
        this.modelBindings = modelBindings;
    }

    /**
     * <p>
     *  This sets the output directory to output generated classes to.
     * </p>
     *
     * @param baseDir the name of the base directory where
     *        all generated classes should be placed
     */
    public void setOutputDirectory(String outputDir) {
        if (outputDir == null) {
            throw new IllegalArgumentException("A generator cannot have a " +
                "null output directory.");
        }

        this.outputDir = new File(outputDir);
        uninitialize();
    }

    public void setModelBindings(ModelBindings modelBindings) {
        if (modelBindings == null) {
            throw new IllegalArgumentException("A generator cannot have a null model hashtable");
        }
        this.modelBindings = modelBindings;
        uninitialize();
    }

    /**
     * <p>
     *  This sets the output directory to output generated classes to.
     * </p>
     *
     * @param outputDir the <code>File</code> (directory) where
     *        all generated classes should be placed
     */
    public void setOutputDirectory(File outputDir) {
        if (outputDir == null) {
            throw new IllegalArgumentException("A generator cannot have a " +
                "null output directory.");
        }

        this.outputDir = outputDir;
        uninitialize();
    }

    /**
     * <p>
     *  This allows a binding to be ignored. This is an intermediary step
     *    towards allowing binding schemas. The name of the binding should
     *    be passed in here, and no property generation for it will occur.
     * </p>
     *
     * @param xmlName the XML local name of the binding to ignore.
     * @param xmlNamespaceURI the XML namespace URI of the binding to ignore. 
     *        The empty string indicates there is no namespace associated 
     *        with the binding.
     */
    public void setIgnoreBinding(String xmlName, String xmlNamespaceURI) {
        ignoredBindings.add(new XMLName(xmlName, xmlNamespaceURI));
    }

    /**
     * <p>
     *  This allows a binding to be ignored. This is an intermediary step 
     *    towards allowing binding schemas. The name of the binding should 
     *    be passed in here, and no property generation for it will occur. This
     *    convenience version supplies an empty string for the namespace URI.
     * </p>
     *
     * @param xmlName the XML local name of the binding to ignore.
     */
    public void setIgnoreBinding(String xmlName) {
        setIgnoreBinding(xmlName, "");
    }

    /**
     * This will determine if a particular binding should be ignored. The
     * binding is identified by its XML local name and its XML namespace URI.
     *
     * @param xmlName the XML local name of the binding.
     * @param xmlNamespaceURI the XML namespace URI associated with the binding. 
     *        An empty string indicates that there is no namespace associated 
     *        with the binding.
     *
     * @return true if and only if the binding should be ignored
     */
    protected boolean isIgnoredBinding(String xmlName, String xmlNamespaceURI) {
        if (ignoredBindings != null) {
            for (Iterator i = ignoredBindings.iterator(); i.hasNext(); ) {
                XMLName bindingName = (XMLName)i.next();
                if ((bindingName.getXMLName().equals(xmlName)) &&
                    (bindingName.getNamespaceURI().equals(xmlNamespaceURI))) {
                    return true;
                }
            }
        }

        return false;
    }

    /**
     * <p>
     *  This will convert from a set of constraints to Java classes.
     *    It accesses those constraints through the
     *    <code>{@link Binding}</code> interface, which provides
     *    them in a representation-independent format, and then
     *    converts them into Java code.
     * </p>
     *
     * @param binder <code>Binder</code> with ability to convert from
     *        constraints to Zeus <code>Binding</code>s.
     * @param root whether this is the root element
     * @throws <code>IOException</code> - when errors in writing to the
     *         supplied <code>Result</code> occur.
     * @throws <code>ZeusException</code> - when errors in class generation
     *         occur.
     */
    public void generate(Binding binding)
        throws IOException, ZeusException {

        collapsedElements.clear();

        if (!isInitialized()) {
            initialize();
        }


        /**
         * Currently, only Containers can be generated (which
         *   includes ContainerProperties), as AtomicProperties
         *   won't result in classes.
         */
        if (binding instanceof Container) {
            Container container = (Container)binding;

            // Ignore Java primitive types (from collapsed elements)
            if (ClassUtils.isJavaPrimitive(binding.getJavaType())) {
                return;
            }

	    /** the GUI classes set */

	    /** Generate the GUI interface */
	    generateGUIInterface(container);

	    /** Generate the GUI class */
	    generatePanelHandler(container);
	    generatePanel(container);

	    /** Generate the node bean class */
	    generateNodeHandler(container);
	    generateNode(container);

        } else  {
            throw new UnsupportedBindingTypeException(binding);
        }
    }

    /**
     * <p>
     *  This will generate the interface for an GUI connected to a listener
     *  for the supplied <code>Container</code>.
     * </p>
     *
     * @param container <code>Container</code> to generate classes from.
     * @throws <code>IOException</code> - when generation errors occur.
     */
    protected void generateGUIInterface(Container container)
	throws IOException {

	// Get a new writer
	String filename = container.getJavaType() + "GUI.java";
	File outputFile = new File(getDir(container), filename);
	FileWriter writer = new FileWriter(outputFile);
	
	// Write out the source file header comment
	writer.write(ExtZeusDefaults.GENERATED_CLASS_COMMENT);
	
	// Add a package statement if necessary.
	writePackageDeclaration(writer,
				container.getJavaInterfacePackage());

        // Write out the implementation declaration
        writeDeclaration(writer, container,
		"interface",
		"GUI",
		this.interfacesPackage+".GUI",
		"");

        writer.write("}\n");
        
        writer.flush();
        writer.close();

    }

    /**
     * <p>
     *  This will generate the visual representation
     *  for the supplied <code>Container</code>.
     * </p>
     *
     * @param container <code>Container</code> to generate classes from.
     * @throws <code>IOException</code> - when generation errors occur.
     */
    protected void generatePanelHandler(Container container)
        throws IOException {

	List panelHandlerImports = new LinkedList();

	// Get a new writer
	String filename = container.getJavaType() + "PanelHandler.java";
	File outputFile = new File(getDir(container), filename);
	FileWriter writer = new FileWriter(outputFile);

	// Write out the source file header comment
	writer.write(ExtZeusDefaults.GENERATED_CLASS_COMMENT);

	// Add a package statement if necessary.
	writePackageDeclaration(writer,
				container.getJavaInterfacePackage());

	panelHandlerImports.add(this.interfacesPackage+".ExtensionImpl");
	panelHandlerImports.add(this.interfacesPackage+".ExtensionManager");

        // Write out import statements
        writeImportStatements(writer, panelHandlerImports);

        // Write out the implementation declaration
        writeDeclaration(writer, container,
		"class",
		"PanelHandler",
		"ExtensionImpl",
		"");

        List properties = container.getProperties();

        // Write out each property
	writer.write("    private ExtensionManager _extManager;\n\n");
	writer.write("    private ");
	writer.write(container.getJavaType());
	writer.write("Panel _panel;\n\n");
	
        // Write class constructor
	writer.write("    public ");
	writer.write(container.getJavaType());
	writer.write("PanelHandler(");
	writer.write(") {\n");
	writer.write("        _panel = new ");
	writer.write(container.getJavaType());
	writer.write("Panel(this);\n");
	writer.write("    }\n\n");

	
	writer.write("    public ");
	writer.write("Object getInstance() {\n");
	writer.write("        _panel.startListening();\n");
	writer.write("        return (_panel);\n");
	writer.write("    }\n\n");

	// Write out the methods involving properties
	writer.write("    public String toString() {\n");
	writer.write("        return(\"Panel\");\n");
	writer.write("    }\n\n");
	
        writer.write("}\n");

        writer.flush();
        writer.close();

    }

    /**
     * <p>
     *  This will generate the visual representation
     *  for the supplied <code>Container</code>. 
     * </p>
     *
     * @param container <code>Container</code> to generate classes from.
     * @throws <code>IOException</code> - when generation errors occur.
     */
    protected void generatePanel(Container container)
        throws IOException {

	List panelImports = new LinkedList();

	// Get a new writer
	String filename = container.getJavaType() + "Panel.java";
	File outputFile = new File(getDir(container), filename);
	FileWriter writer = new FileWriter(outputFile);

	// Write out the source file header comment
	writer.write(ExtZeusDefaults.GENERATED_CLASS_COMMENT);

	// Add a package statement if necessary.
	writePackageDeclaration(writer,
				container.getJavaInterfacePackage());

	panelImports.add("java.awt.Component");
	panelImports.add("java.awt.BorderLayout");
	panelImports.add("java.awt.event.ActionEvent");
	panelImports.add("java.awt.event.ActionListener");
	panelImports.add("java.util.List");
	panelImports.add("javax.swing.JButton");
	panelImports.add("javax.swing.JLabel");
	panelImports.add("javax.swing.JPanel");

	panelImports.add(this.interfacesPackage+".GUI");

        // Write out import statements
        writeImportStatements(writer, panelImports);

        // Write out the implementation declaration
        writeDeclaration(writer, container,
		"class",
		"Panel",
		"JPanel",
		container.getJavaType()+"GUI,ActionListener");

        List properties = container.getProperties();

        // Write out each property
        writeGUIPanelProperties(writer,
					container.getJavaType(),
					properties);

        // Write class constructor
        writeGUIPanelConstructor(writer,
                                       container.getJavaType(),
				       properties);

	// Write out the methods involving properties
	writeGUIPanelMethods(writer, container.getJavaType(), properties, container.getJavaInterfacePackage());

        writer.write("}\n");

        writer.flush();
        writer.close();

    }

    /**
     * <p>
     *  This will generate the visual representation
     *  for the supplied <code>Container</code>.
     * </p>
     *
     * @param container <code>Container</code> to generate classes from.
     * @throws <code>IOException</code> - when generation errors occur.
     */
    protected void generateNodeHandler(Container container)
        throws IOException {

	List panelHandlerImports = new LinkedList();

	// Get a new writer
	String filename = container.getJavaType() + "NodeHandler.java";
	File outputFile = new File(getDir(container), filename);
	FileWriter writer = new FileWriter(outputFile);
	
	// Write out the source file header comment
	writer.write(ExtZeusDefaults.GENERATED_CLASS_COMMENT);

	// Add a package statement if necessary.
	writePackageDeclaration(writer,
				container.getJavaInterfacePackage());

	panelHandlerImports.add(this.interfacesPackage+".ExtensionImpl");
	panelHandlerImports.add(this.interfacesPackage+".ExtensionManager");

        // Write out import statements
        writeImportStatements(writer, panelHandlerImports);

        // Write out the implementation declaration
        writeDeclaration(writer, container,
		"class",
		"NodeHandler",
		"ExtensionImpl",
		"");

        List properties = container.getProperties();

        // Write out each property
	writer.write("    private ExtensionManager _extManager;\n\n");
	writer.write("    private ");
	writer.write(container.getJavaType());
	writer.write("Node _node;\n\n");
	
        // Write class constructor
	writer.write("    public ");
	writer.write(container.getJavaType());
	writer.write("NodeHandler(");
	writer.write(") {\n");
	writer.write("        _node = new ");
	writer.write(container.getJavaType());
	writer.write("Node(this);\n");
	writer.write("    }\n\n");


	writer.write("    public ");
	writer.write("Object getInstance() {\n");
	writer.write("        _node.startListening();\n");
	writer.write("        return (_node);\n");
	writer.write("    }\n\n");

	// Write out the methods involving properties
	writer.write("    public String toString() {\n");
	writer.write("        return(\"Node\");\n");
	writer.write("    }\n\n");
	
        writer.write("}\n");
        
        writer.flush();
        writer.close();

    }

    /**
     * <p>
     *  This will generate the visual node representation 
     *  for the supplied <code>Container</code>. 
     * </p>
     *
     * @param container <code>Container</code> to generate classes from.
     * @throws <code>IOException</code> - when generation errors occur.
     */
    protected void generateNode(Container container) 
        throws IOException {

	// Get a new writer
	String filename = container.getJavaType() + "Node.java";
	File outputFile = new File(getDir(container), filename);
	FileWriter writer = new FileWriter(outputFile);
	
	// Write out the source file header comment
	writer.write(ExtZeusDefaults.GENERATED_CLASS_COMMENT);

	// Add a package statement if necessary.
	writePackageDeclaration(writer,
				container.getJavaInterfacePackage());

        writeDeclaration(writer, container,
		"class",
		"Node",
		"javax.swing.tree.DefaultMutableTreeNode",
		container.getJavaType()+"GUI");

	writer.write("    private ");
	writer.write(container.getJavaType());
	writer.write("NodeHandler _handler;\n\n");
	writer.write("    private ");
	writer.write(container.getJavaType());
	writer.write("Listener _listener;\n\n");

        writer.write("    public ");
        writer.write(container.getJavaType());
        writer.write("Node(");
	writer.write(container.getJavaType());
	writer.write("NodeHandler handler) {\n");
	writer.write("        _handler = handler ;\n");

	writer.write("    }\n\n");

	writer.write("    public void startListening() {\n");
	writer.write("        _listener = new ");
	writer.write(container.getJavaType());
	writer.write("ListenerImpl( (");
	writer.write(container.getJavaType());
	writer.write("GUI) this);\n");
	writer.write("    }\n\n");

	writer.write("    public ");
	writer.write(this.interfacesPackage);
	writer.write(".ExtensionManager getManager() {\n");
	writer.write("        return (_handler.getManager());\n");
	writer.write("    }\n\n");

	writer.write("    public void refresh() {\n");
	writer.write("        removeAllChildren();\n");

        List properties = container.getProperties();

	for (Iterator i = properties.iterator(); i.hasNext(); ) {
            Property property = (Property)i.next();
            if (property.getModifier().get(Property.SOURCE_ELEMENT))
	    {
		if (property.isCollection())
		{
		    writer.write("        java.util.List ");
		    writer.write(property.getJavaName());
		    writer.write("List = ((");
		    writer.write(container.getJavaType());
		    writer.write("Bean)_handler.getManager().getSource()).get");
		    writer.write(CapitalizationUtils.initialUpper(property.getJavaName()));
		    writer.write("List();\n");
		    writer.write("        for (java.util.Iterator i = ");
		    writer.write(property.getJavaName());
		    writer.write("List.iterator(); i.hasNext(); ) {\n");
		    writer.write("            ");
		    writer.write(this.interfacesPackage);
		    writer.write(".Bean child = (");
		    writer.write(this.interfacesPackage);
		    writer.write(".Bean)i.next();\n");
		    writer.write("            this.add ( (javax.swing.tree.DefaultMutableTreeNode) (child.getExtensionManager().getExtensionByName(\"Node\").getInstance()) );\n");
		    writer.write("        }\n");
		}
		else
		{
		    writer.write("        if ( ((");
		    writer.write(container.getJavaType());
		    writer.write("Bean)_handler.getManager().getSource()).get");
		    writer.write(CapitalizationUtils.initialUpper(property.getJavaName()));
		    writer.write("() != null ) \n");
		    writer.write("            this.add(   (javax.swing.tree.DefaultMutableTreeNode) ( (");
		    writer.write(CapitalizationUtils.initialUpper(property.getXMLName()));
		    writer.write("Bean) (((");
		    writer.write(container.getJavaType());
		    writer.write("Bean)_handler.getManager().getSource()).get");
		    writer.write(CapitalizationUtils.initialUpper(property.getJavaName()));
		    writer.write("()) ).getExtensionManager().getExtensionByName(\"Node\").getInstance()    ) ;\n");
		}
	    }
	}

	writer.write("    }\n\n");

	writer.write("    public void validateChanges() {\n");
	writer.write("    }\n");

	writer.write("    public String toString() {\n");
	writer.write("        return(\"");
	writer.write(container.getJavaType());
	writer.write("\");\n");
	writer.write("    }\n");

        writer.write("    public void removeChild(String child_class, ");
        writer.write(interfacesPackage);
        writer.write(".Bean child_to_remove) {\n");
	for (Iterator i = properties.iterator(); i.hasNext(); ) {
            Property property = (Property)i.next();
            writer.write("        if (child_class.equals(\"");
            writer.write(property.getJavaName());
            writer.write("\"))\n");
            writer.write("            ((");
            writer.write(container.getJavaType());
            writer.write("Operator) ((");
            writer.write(container.getJavaType());
            writer.write("Bean) this.getManager().getSource() ).getExtensionManager().getExtensionByName(\"Operator\").getInstance() ).operate(\"remove");
            writer.write(CapitalizationUtils.initialUpper(property.getJavaName()));
            writer.write("\", child_to_remove);\n");
        }
        writer.write("    }\n\n");

        writer.write("}\n");

        writer.flush();
        writer.close();

    }

    /**
     * <p>
     *  This writes out the package declaration for this class, if needed.
     * </p>
     *
     * @param writer <code>Writer</code> to perform output to
     * @param packageName the package of this class
     * @throws <code>IOException</code> - when errors in output occur.
     */
    protected void writePackageDeclaration(Writer writer, String packageName)
        throws IOException {
            
        // Generate a package statement if necessary.
        if (!packageName.equals("")) {
            writer.write("package");
            writer.write(" ");
            writer.write(packageName);
            writer.write(";\n\n");
        }
    }
    
    /**
     * <p>
     *  This will take the supplied list of class names, and add import
     *    statements for each of the classes in the list.
     * </p>
     *
     * @param writer <code>Writer</code> to output to
     * @param importClasses <code>List</code> of classes to add imports for
     * @throws <code>IOException</code> - when errors in output occur.
     */
    protected void writeImportStatements(Writer writer, 
                                                  List importClasses)
        throws IOException {

        // Write out imports
        if (importClasses.size() > 0) {
            writer.write("// Local Interface Import Statements\n");
        }
        for (Iterator i = importClasses.iterator(); i.hasNext(); ) {
            writer.write("import ");
            writer.write((String)i.next());
            writer.write(";\n");
        }
        if (importClasses.size() > 0) {
            writer.write("\n");
        }
    }
    
    /**
     * <p>
     *  This writes out the interface bean declaration (the class header).
     * </p>
     *
     * @param writer <code>Writer</code> to output to
     * @param container the container to write the interface for.
     * @throws <code>IOException</code> - when errors in output occur.
     */
    protected void writeDeclaration(Writer writer,
                                             Container container,
					     String modifier,
					     String suffix,
					     String extended,
					     String implemented)
        throws IOException {

        writer.write("public ");
	writer.write(modifier);
	writer.write(" ");
        writer.write(container.getJavaType());
	writer.write(suffix);
	if (!extended.equals("")) {
	    writer.write(" extends ");
	    writer.write(extended);
	}

	if (!implemented.equals("")) {
	    writer.write(" implements ");
	    writer.write(implemented);
	}

	writer.write(" {\n\n");
    }

    /**
     * <p>
     *  This writes out the properties for an GUI Bean as variables.
     * </p>
     *
     * @param writer <code>Writer</code> to output to
     * @param properties <code>List</code> of properties to write out as
     *        member variables
     * @throws <code>IOException</code> - when errors in output occur.
     */
    protected void writeGUIPanelProperties(Writer writer, String className, List properties)
        throws IOException {

	writer.write("    private ");
	writer.write(className);
	writer.write("PanelHandler _handler;\n");
	writer.write("    private ");
	writer.write(className);
	writer.write("Listener _listener;\n\n");
	writer.write("    private JPanel contained_attr = new JPanel();\n");
        writer.write("    private JPanel contained_children = new JPanel();\n");
        writer.write("    private javax.swing.BoxLayout layout1 = new javax.swing.BoxLayout(contained_attr, javax.swing.BoxLayout.Y_AXIS);\n");
        writer.write("    private javax.swing.BoxLayout layout2 = new javax.swing.BoxLayout(contained_children, javax.swing.BoxLayout.Y_AXIS);\n");
        writer.write("    javax.swing.JPopupMenu operations_menu = new javax.swing.JPopupMenu();\n");
        for (Iterator i = properties.iterator(); i.hasNext(); ) {
            Property property = (Property)i.next();
            if (isIgnoredBinding(property.getXMLName(),
                                 property.getXMLNamespaceURI())) {
                continue;
            }
            writeGUIPanelProperty(writer, property, className);
        }

    }

    /**
     * <p>
     *  This will take a <code>Property</code> and generate the code
     *    for that property as a member variable SPECIFIC to GUIs.
     * </p>
     *
     * @param writer <code>Writer</code> to write variable declaration to.
     * @param property <code>Property</code> to generate code for.
     * @throws <code>IOException</code> - when errors in writing to the
     *         supplied <code>Result</code> occur.
     */
    protected void writeGUIPanelProperty(Writer writer, Property property, String className)
        throws IOException {

        /**
         * There is a conditional here, although all that ever should
         *   get to here are AtomicProperty objects. But just in case...
         */
        if ((property instanceof AtomicProperty) &&
            (!property.getXMLName().startsWith("xmlns")) &&
	    (	(property.getModifier().get(Property.SOURCE_ATTLIST))) ||
		(ClassUtils.isJavaPrimitive(property.getJavaType())
                ||
                ( property.getModifier().get(Property.SOURCE_ELEMENT)
                  && this.modelBindings.getMappingFor(property.getXMLName(),"#","inparent").equals("true"))
                )
	    )
	{
            if (!property.isCollection())
            {
                AtomicProperty atomicProperty = (AtomicProperty)property;
                writer.write("    private ");
                if (property.getModifier().get(Property.SOURCE_ELEMENT))
                {
                    writer.write(this.interfacesPackage);
                    writer.write(".GUI");
                }
                else
                    writer.write( this.modelBindings.getMappingFor(CapitalizationUtils.allLower(className),property.getXMLName(),"javaclass"));
                writer.write(" visualField");
                writer.write(CapitalizationUtils.initialUpper(atomicProperty.getJavaName()));
	        writer.write(";\n");
                writer.write("    private JPanel ");
                writer.write(property.getJavaName());
                writer.write("Panel = new JPanel(");
                writer.write("new java.awt.GridLayout(0,1)");
                writer.write(");\n\n");
            }
        }
    }

    /**
     * <p>
     *  This writes out the constructor for an GUI Bean class
     * </p>
     *
     * @param writer <code>Writer</code> to output to
     * @param className the name of the Bean class
     * @throws <code>IOException</code> - when errors in output occur.
     */
    protected void writeGUIPanelConstructor(Writer writer,
                                                  String className,
						  List properties)
        throws IOException {

        BitSet modifiers = null;

        writer.write("    public ");
        writer.write(className);
        writer.write("Panel(");
	writer.write(className);
	writer.write("PanelHandler handler) {\n");

        if (this.modelBindings.getMappingFor(CapitalizationUtils.allLower(className),"#","displaylabel").equals("true"))
        {
            writer.write("        this.setBorder(javax.swing.BorderFactory.createTitledBorder(\"");
            writer.write(className);
            writer.write("\"));\n");
        }
	writer.write("        this.setLayout(new BorderLayout());\n");
	writer.write("        _handler = handler ;\n");
	writer.write("        this.add(contained_attr, BorderLayout.CENTER);\n");
	writer.write("        this.add(contained_children, BorderLayout.SOUTH);\n");
        writer.write("        contained_attr.setLayout(layout1);\n");
        writer.write("        contained_children.setLayout(layout2);\n");
	for (Iterator i = properties.iterator(); i.hasNext(); )
        {
	    Property property = (Property)i.next();
        if (property.getModifier().get(Property.SOURCE_ELEMENT)
                && this.modelBindings.getMappingFor(property.getXMLName(),"#","inparent").equals("true")
                && property.isCollection())
                    {
                        writer.write("        javax.swing.JMenuItem ");
                        writer.write(property.getJavaName());
                        writer.write("_adder = new javax.swing.JMenuItem(\"+ ");
                        writer.write(property.getJavaName());
                        writer.write("\");\n");
                        writer.write("        ");
                        writer.write(property.getJavaName());
                        writer.write("_adder.setActionCommand(\"add");
                        writer.write(CapitalizationUtils.initialUpper(property.getJavaName()));
                        writer.write("\");\n");
                        writer.write("        ");
                        writer.write(property.getJavaName());
                        writer.write("_adder.addActionListener(this);\n");
                        writer.write("        operations_menu.add(");
                        writer.write(property.getJavaName());
                        writer.write("_adder);\n");
                    }
        if (property.getModifier().get(Property.SOURCE_ELEMENT)
                && this.modelBindings.getMappingFor(property.getXMLName(),"#","inparent").equals("true")
                && !property.isCollection())
                    {
                     writer.write("        javax.swing.JMenuItem ");
                        writer.write(property.getJavaName());
                        writer.write("_setter = new javax.swing.JMenuItem(\"set/new ");
                        writer.write(property.getJavaName());
                        writer.write("\");\n");
                        writer.write("        ");
                        writer.write(property.getJavaName());
                        writer.write("_setter.setActionCommand(\"set");
                        writer.write(CapitalizationUtils.initialUpper(property.getJavaName()));
                        writer.write("\");\n");
                        writer.write("        ");
                        writer.write(property.getJavaName());
                        writer.write("_setter.addActionListener(this);\n");
                        writer.write("        operations_menu.add(");
                        writer.write(property.getJavaName());
                        writer.write("_setter);\n");
                    }
	}

        writer.write("        javax.swing.JPanel north_panel = new javax.swing.JPanel(new java.awt.BorderLayout());\n");
        writer.write("        javax.swing.JPanel enlarger = new javax.swing.JPanel();\n");
        writer.write("        enlarger.setPreferredSize(new java.awt.Dimension(330,15));\n");
        writer.write("        this.add(north_panel, BorderLayout.NORTH);\n");
        writer.write("        north_panel.add(enlarger, BorderLayout.CENTER);\n");
        writer.write("        if ( operations_menu.getComponentCount() >0)\n");
        writer.write("        {\n");
	writer.write("            javax.swing.JButton action_button = new javax.swing.JButton();\n");
        writer.write("            action_button.setPreferredSize(new java.awt.Dimension(15,15));\n");
        writer.write("            action_button.setBackground(java.awt.Color.red);\n");
        writer.write("            action_button.setActionCommand(\"display_menu\");\n");
        writer.write("            action_button.addActionListener(this);\n");
        writer.write("            north_panel.add(action_button, BorderLayout.WEST);\n");
        writer.write("        }\n");
	writer.write("        javax.swing.JButton remove_button = new javax.swing.JButton();\n");
        writer.write("        remove_button.setPreferredSize(new java.awt.Dimension(15,15));\n");
        writer.write("        remove_button.setBackground(java.awt.Color.black);\n");
        writer.write("        remove_button.setActionCommand(\"remove\");\n");
        writer.write("        remove_button.addActionListener(this);\n");
        writer.write("        north_panel.add(remove_button, BorderLayout.EAST);\n");


        for (Iterator i = properties.iterator(); i.hasNext(); ) {
            Property property = (Property)i.next();
	    if (property.getModifier().get(Property.SOURCE_ATTLIST)
                ||
		(ClassUtils.isJavaPrimitive(property.getJavaType()))
                ||
                ( property.getModifier().get(Property.SOURCE_ELEMENT)
                  && this.modelBindings.getMappingFor(property.getXMLName(),"#","inparent").equals("true"))
                )
	    {
                if (!property.isCollection())
                    if (!property.getModifier().get(Property.SOURCE_ELEMENT))
                    {
                        writer.write("        visualField");
                        writer.write(CapitalizationUtils.initialUpper(property.getJavaName()));
                        writer.write(" = new ");
                        writer.write(this.modelBindings.getMappingFor(CapitalizationUtils.allLower(className),property.getXMLName(),"javaclass"));
                        writer.write("();\n");
                    }
	    }
	}

	for (Iterator i = properties.iterator(); i.hasNext(); ) {
	    Property property = (Property)i.next();
	    if (property.getModifier().get(Property.SOURCE_ATTLIST) ||
		(ClassUtils.isJavaPrimitive(property.getJavaType()))
                )
	    {

                if (!property.isCollection())
                {
                    writer.write("        ");
                    writer.write(property.getJavaName());
                    writer.write("Panel.add( ");
                    writer.write("visualField");
                    writer.write(CapitalizationUtils.initialUpper(property.getJavaName()));
                    writer.write(");\n");
                    if (!property.getXMLName().equals(ZeusDefaults.PCDATA_XML_NAME)
                        && this.modelBindings.getMappingFor(CapitalizationUtils.allLower(className),property.getJavaName(),"displaylabel").equals("true")
                        )
                    {
                        writer.write("        ");
                        writer.write(property.getJavaName());
                        writer.write("Panel.setBorder(javax.swing.BorderFactory.createTitledBorder(\"");
                        writer.write(property.getJavaName());
                        writer.write("\"));\n");
                    }
		    writer.write("        contained_attr.add(");
                    writer.write(property.getJavaName());
                    writer.write("Panel);\n");
                }
	    }
        }
	writer.write("        this.repaint();\n");
        writer.write("    }\n\n");
    }

    /**
     * <p>
     *  This writes out the methods involving properties for the GUI Bean
     * </p>
     *
     * @param writer <code>Writer</code> to output to
     * @param properties <code>List</code> of properties to write out as
     *        member variables
     * @throws <code>IOException</code> - when errors in output occur.
     */
    protected void writeGUIPanelMethods(Writer writer,
                                                  String className,
                                                  List properties,
						  String packageName)

        throws IOException {

        writer.write("    public void startListening() {\n");
        writer.write("        refresh();\n");
        writer.write("        _listener = new ");
        writer.write(className);
        writer.write("ListenerImpl( (");
        writer.write(className);
        writer.write("GUI) this);\n");
        writer.write("    }\n\n");


        writer.write("    public void actionPerformed(ActionEvent cliquage) {\n\n");
        writer.write("        String command = cliquage.getActionCommand();\n");
        writer.write("        if ( command.equals(\"display_menu\"))\n");
        writer.write("            showMenu( ((javax.swing.JButton)cliquage.getSource() ).getLocation());\n");
        writer.write("        if ( command.equals(\"remove\")) {\n");
        writer.write("            ((");
        writer.write(interfacesPackage);
        writer.write(".GUI) this.getParent().getParent() ).removeChild(\"");
        writer.write(className);
        writer.write("\", (");
        writer.write(interfacesPackage);
        writer.write(".Bean) this.getManager().getSource());\n");
        writer.write("        }\n");
	for (Iterator i = properties.iterator(); i.hasNext(); )
        {
            Property property = (Property)i.next();
            if (( property.getModifier().get(Property.SOURCE_ELEMENT)
                  && this.modelBindings.getMappingFor(property.getXMLName(),"#","inparent").equals("true"))
                &&
                property.isCollection())
            {
                writer.write("        if ( command.equals(\"add");
                writer.write(CapitalizationUtils.initialUpper(property.getJavaName()));
                writer.write("\")) {\n");
                writer.write("            this.validateChanges();\n");
                writer.write("            ((");
                writer.write(className);
                writer.write("Operator) ((");
                writer.write(className);
                writer.write("Bean) this.getManager().getSource() ).getExtensionManager().getExtensionByName(\"Operator\").getInstance() ).operate(\"add");
                writer.write(CapitalizationUtils.initialUpper(property.getJavaName()));
                writer.write("\",null);\n");
                writer.write("            this.my_validate();\n");
                writer.write("        }\n");
            }
            if (( property.getModifier().get(Property.SOURCE_ELEMENT)
                  && this.modelBindings.getMappingFor(property.getXMLName(),"#","inparent").equals("true"))
                &&
                !property.isCollection())
            {
                writer.write("        if ( command.equals(\"set");
                writer.write(CapitalizationUtils.initialUpper(property.getJavaName()));
                writer.write("\")) {\n");
                writer.write("            this.validateChanges();\n");
                writer.write("            ((");
                writer.write(className);
                writer.write("Operator) ((");
                writer.write(className);
                writer.write("Bean) this.getManager().getSource() ).getExtensionManager().getExtensionByName(\"Operator\").getInstance() ).operate(\"set");
                writer.write(CapitalizationUtils.initialUpper(property.getJavaName()));
                writer.write("\",null);\n");
                writer.write("            this.my_validate();\n");
                writer.write("        }\n");
                writer.write("        if ( command.equals(\"unset");
                writer.write(CapitalizationUtils.initialUpper(property.getJavaName()));
                writer.write("\")) {\n");
                writer.write("            this.validateChanges();\n");
                writer.write("            ((");
                writer.write(className);
                writer.write("Operator) ((");
                writer.write(className);
                writer.write("Bean) this.getManager().getSource() ).getExtensionManager().getExtensionByName(\"Operator\").getInstance() ).operate(\"unset");
                writer.write(CapitalizationUtils.initialUpper(property.getJavaName()));
                writer.write("\",null);\n");
                writer.write("            this.my_validate();\n");
                writer.write("        }\n");

            }
        }
        writer.write("    }\n\n");


        writer.write("    public void validateChanges() {\n\n");
        writer.write("            for (int iterator=0 ; iterator<contained_children.getComponentCount(); iterator++) {\n");
        writer.write("                ");
        writer.write(this.interfacesPackage);
        writer.write(".GUI");
        writer.write(" thisIteratedField = (");
        writer.write(this.interfacesPackage);
        writer.write(".GUI");
        writer.write(") contained_children.getComponent(iterator);\n");
        writer.write("                thisIteratedField.validateChanges();\n");
        writer.write("            }\n");

	for (Iterator i = properties.iterator(); i.hasNext(); ) {
            Property property = (Property)i.next();
	    if ( (! isIgnoredBinding(property.getXMLName(),
                                 property.getXMLNamespaceURI())) &&
		(!property.getXMLName().startsWith("xmlns")) &&
		(!property.getModifier().get(Property.MUTABILITY_FINAL)) &&
		( (property.getModifier().get(Property.SOURCE_ATTLIST)) ||
		(ClassUtils.isJavaPrimitive(property.getJavaType()))
                ||
                ( property.getModifier().get(Property.SOURCE_ELEMENT)
                  && this.modelBindings.getMappingFor(property.getXMLName(),"#","inparent").equals("true"))
                )
		)
	    {
                if (!property.isCollection())
                {
                    if (property.getModifier().get(Property.SOURCE_ELEMENT))
                    {
                        writer.write("        if( (");
                        writer.write(interfacesPackage);
                        writer.write(".Bean)( (");
                        writer.write(className);
                        writer.write("Bean)_handler.getManager().getSource()).get");
                        writer.write(CapitalizationUtils.initialUpper(property.getJavaName()));
                        writer.write("() != null)\n");
                        writer.write("    ");
                    }
                    writer.write("            ((");
                    writer.write(className);
                    writer.write("Bean)_handler.getManager().getSource()).");
                    writer.write("set");
		    writer.write(CapitalizationUtils.initialUpper(property.getJavaName()));
                    writer.write("( ");
                    if (property.getModifier().get(Property.SOURCE_ELEMENT))
                    {
                        writer.write("(");
                        writer.write(CapitalizationUtils.initialUpper(property.getJavaType()));
                        writer.write("Bean)");
                    }writer.write("(visualField");
                    writer.write(CapitalizationUtils.initialUpper(property.getJavaName()));
                    writer.write(".");
                    if (property.getModifier().get(Property.SOURCE_ELEMENT))
                        writer.write("getManager().getSource");
                    else
                        writer.write(this.modelBindings.getMappingFor(CapitalizationUtils.allLower(className),property.getXMLName(),"retrievemethod"));
                    writer.write("() ));\n");
                }
	    }
	}
	writer.write("\n        _handler.getManager().getSource().notifyListeners();\n");
	writer.write("    }\n\n");

	writer.write("    public void refresh() {\n\n");

        writer.write("        contained_children.removeAll();\n");
	for (Iterator i = properties.iterator(); i.hasNext(); ) {
            Property property = (Property)i.next();
	    if (property.getModifier().get(Property.SOURCE_ATTLIST) ||
		(ClassUtils.isJavaPrimitive(property.getJavaType()))
                ||
                ( property.getModifier().get(Property.SOURCE_ELEMENT)
                  && this.modelBindings.getMappingFor(property.getXMLName(),"#","inparent").equals("true"))
                )
	    {
                if (property.isCollection())
                {
                    writer.write("            for (java.util.Iterator thisCollectionIterator = ((");
                    writer.write(className);
                    writer.write("Bean)_handler.getManager().getSource()).get");
                    writer.write(CapitalizationUtils.initialUpper(property.getJavaName()));
                    writer.write("List().iterator(); thisCollectionIterator.hasNext(); ) {\n");
                    writer.write("                ");
                    if (property.getModifier().get(Property.SOURCE_ELEMENT))
                    {
                        writer.write(interfacesPackage);
                        writer.write(".Bean ");
                    }
                    else
                        writer.write("String ");
                    writer.write("thisIteratedProperty = (");
                    if (property.getModifier().get(Property.SOURCE_ELEMENT))
                    {
                        writer.write(interfacesPackage);
                        writer.write(".Bean");
                    }
                    else
                        writer.write("String");
                    writer.write(") thisCollectionIterator.next();\n");
                    writer.write("                ");
                    if (property.getModifier().get(Property.SOURCE_ELEMENT))
                    {
                        writer.write(this.interfacesPackage);
                        writer.write(".GUI");
                    }
                    else
                        writer.write(this.modelBindings.getMappingFor(CapitalizationUtils.allLower(className),property.getXMLName(),"javaclass"));
                    writer.write(" correspondingIteratedField = ");
                    if (property.getModifier().get(Property.SOURCE_ELEMENT))
                        writer.write("(GUI) (thisIteratedProperty.getExtensionManager().getExtensionByName(\"Panel\").getInstance());\n");
                    else{
                        writer.write("new ");
                        writer.write(this.modelBindings.getMappingFor(CapitalizationUtils.allLower(className),property.getXMLName(),"javaclass"));
                        writer.write("();\n");
                    }
                    if (!property.getModifier().get(Property.SOURCE_ELEMENT))
                    {
                        writer.write("                correspondingIteratedField.");
                        writer.write(this.modelBindings.getMappingFor(CapitalizationUtils.allLower(className),property.getXMLName(),"feedmethod"));
                        writer.write("(thisIteratedProperty);\n");
                    }
                    writer.write("                contained_children.add( (");
                    writer.write(CapitalizationUtils.initialUpper(property.getJavaType()));
                    writer.write("Panel)correspondingIteratedField);\n");
                    writer.write("            }\n");
                } else {
                    if (property.getModifier().get(Property.SOURCE_ELEMENT))
                    {
                        writer.write("        if( (");
                        writer.write(interfacesPackage);
                        writer.write(".Bean)( (");
                        writer.write(className);
                        writer.write("Bean)_handler.getManager().getSource()).get");
                        writer.write(CapitalizationUtils.initialUpper(property.getJavaName()));
                        writer.write("() != null) {\n");
                     writer.write("        visualField");
                        writer.write(CapitalizationUtils.initialUpper(property.getJavaName()));
                        writer.write("=(GUI) ((");
                        writer.write(interfacesPackage);
                        writer.write(".Bean)( (");
                        writer.write(className);
                        writer.write("Bean)_handler.getManager().getSource()).get");
                        writer.write(CapitalizationUtils.initialUpper(property.getJavaName()));
                        writer.write("()).");
                        writer.write("getExtensionManager().getExtensionByName(\"Panel\").getInstance()");
                        writer.write(";\n");

                        writer.write("        contained_children.add(");
                        writer.write("("+CapitalizationUtils.initialUpper(property.getJavaType()));
                        writer.write("Panel)");
                        writer.write("visualField");
                        writer.write(CapitalizationUtils.initialUpper(property.getJavaName()));
                        writer.write(");\n");
                        writer.write("    }\n");
                    }else{
                     writer.write("        visualField");
                        writer.write(CapitalizationUtils.initialUpper(property.getJavaName()));
                        writer.write(".");
                        writer.write(this.modelBindings.getMappingFor(CapitalizationUtils.allLower(className),property.getXMLName(),"feedmethod"));
                        writer.write("(((");
		        writer.write(className);
		        writer.write("Bean)_handler.getManager().getSource()).get");
                        writer.write(CapitalizationUtils.initialUpper(property.getJavaName()));
                        writer.write("());\n");
                    }
                }
	    }
	}
	writer.write("    }\n\n");

	writer.write("    public ");
	writer.write(this.interfacesPackage);
	writer.write(".ExtensionManager getManager() {\n");
	writer.write("        return (_handler.getManager());\n");
	writer.write("    }\n\n");

        writer.write("    public void removeChild(String child_class, ");
        writer.write(interfacesPackage);
        writer.write(".Bean child_to_remove) {\n");
	for (Iterator i = properties.iterator(); i.hasNext(); ) {
            Property property = (Property)i.next();
            if (property.getModifier().get(Property.SOURCE_ELEMENT))
            {
                writer.write("        if (child_class.equals(\"");
                writer.write(CapitalizationUtils.initialUpper(property.getJavaName()));
                writer.write("\"))\n");
                writer.write("            ((");
                writer.write(className);
                writer.write("Operator) ((");
                writer.write(className);
                writer.write("Bean) this.getManager().getSource() ).getExtensionManager().getExtensionByName(\"Operator\").getInstance() ).operate(\"");
                if (property.isCollection())
                    writer.write("remove");
                else
                    writer.write("unset");
                writer.write(CapitalizationUtils.initialUpper(property.getJavaName()));
                writer.write("\", child_to_remove);\n");
            }
        }
        writer.write("        this.my_validate();\n");

        writer.write("    }\n\n");

        writer.write("    public void my_validate() {\n");
        writer.write("        this.setVisible(false);\n");
        writer.write("        this.validate();\n");
        writer.write("        this.setVisible(true);\n");
        writer.write("        ((javax.swing.JFrame) this.getRootPane().getParent() ).pack();\n");
        writer.write("    }\n\n");

        writer.write("    protected void showMenu(java.awt.Point p){\n");
        writer.write("        operations_menu.show((java.awt.Component)this, (int)p.getX() + 30, (int)p.getY() + 30);\n");
        writer.write("    }\n\n");

    }

    /**
     * <p>
     *  This will initialize this instance of <code>Generator</code>,
     *    handling tasks like directory creation.
     * </p>
     *
     * @throws <code>IllegalOutputDirectoryException</code> - when an illegal
     *         output directory has been supplied to this <code>Generator</code>
     */
    private void initialize() throws IllegalOutputDirectoryException {
        // Determine the output directory
        if (outputDir == null) {
            outputDir = new File(".");
        }

        // Ensure the output directory exists and is a directory
        if ((outputDir.exists()) && (!outputDir.isDirectory())) {
            throw new IllegalOutputDirectoryException(
                outputDir.getAbsolutePath());
        }

        // Create the directory if it does not exist
        if (!outputDir.exists()) {
            outputDir.mkdir();
        }

        initialized = true;
    }

    /**
     * <p>
     *  This will reset the initialization state of this <code>Generator</code>
     *    instance to <code>false</code>.
     * </p>
     */
    private void uninitialize() {
        initialized = false;
    }

    /**
     * <p>
     *  This will indicate if this <code>Generator</code> instance has been
     *    initialized.
     * </p>
     *
     * @return <code>boolean</code> - whether this instance is initialized.
     */
    private boolean isInitialized() {
        return initialized;
    }
    
    /**
     * <p>
     *  This will determine the directory to use for generation
     *    based on the packages of a <code>{@link Container}</code>.
     * </p>
     *
     * @param container the container to obtain the directory for.
     * @return <code>File</code> - the directory for generation.
     */
    private File getDir(Container container) {
        StringBuffer dirName = new StringBuffer();
        String thisPackage = container.getJavaInterfacePackage();

        for (int i=0, len=thisPackage.length(); i<len; i++) {
            char ch = thisPackage.charAt(i);
            if (ch != '.') {
                dirName.append(ch);
            } else {
                dirName.append(System.getProperty("file.separator"));
            }
        }
        
        File dir = 
            new File(outputDir, dirName.toString());
        
        // Create the directory if it does not exist
        if (!dir.exists()) {
            dir.mkdirs();
        }

        return dir;
    }
}
