/*====================================================================

OpenCCM: The Open CORBA Component Model Platform
Copyright (C) 2000-2002 USTL - LIFL - GOAL
Contact: openccm@objectweb.org

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
USA

Initial developer(s): Christophe Contreras
Contributor(s): ___________________________________________________.

====================================================================*/

package org.objectweb.zeusx;

import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Writer;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Vector;
import java.util.BitSet;

// Zeus imports
import org.enhydra.zeus.Binding;
import org.enhydra.zeus.Generator;
import org.enhydra.zeus.ZeusDefaults;
import org.enhydra.zeus.ZeusException;
import org.enhydra.zeus.binding.AtomicProperty;
import org.enhydra.zeus.binding.BaseBinding;
import org.enhydra.zeus.binding.Container;
import org.enhydra.zeus.binding.Property;
import org.enhydra.zeus.util.CapitalizationUtils;
import org.enhydra.zeus.util.ClassUtils;
import org.enhydra.zeus.util.NamingUtils;
import org.enhydra.zeus.util.XMLName;

//imports needed for dtd reading
import org.enhydra.zeus.Transformer;
import org.enhydra.zeus.binder.DTDBinder;
import org.enhydra.zeus.source.DTDSource;
import org.enhydra.zeus.source.StreamDTDSource;
import org.enhydra.zeus.transform.DefaultsTransformer;

import org.enhydra.zeus.generator.*;


/**
 * <p>
 *  <code>{@link Generator}</code> is the portion of Zeus that will convert
 *    a set of constraints, visible through a <code>{@link Binding}</code>,
 *    and output them as generated Java classes (to a
 *    <code>File</code> [directory]).
 * </p><p>
 *  <code>ObjectWebOperatorGenerator</code> is the ObjectWeb XmlTool additional
 *    <code>Generator</code> and takes in either a single <code>Binding</code>
 *    or a <code>List</code> of them, and directly generates Java classes.
 * </p>
 *
 * @author Christophe Contreras
 */
public class ObjectWebOperatorGenerator implements Generator {

    /** A list of bindings to ignore */
    protected List ignoredBindings;

    /** This instances's collapsed elements */
    protected List collapsedElements;

    /** Base directory for generated files. */
    protected File outputDir;

    /** the package where we'll find interfaces. */
    protected String interfacesPackage;

    /** the directory where we'll find models. */
    protected String modelsDirectory;

    protected ModelBindings modelBindings;

    /** Keeps track of whether the Generator needs to be reinitialized. */
    protected boolean initialized;

    /**
     * <p>
     *  This will construct a new instance of <code>ObjectWebOperatorGenerator</code>.
     * </p>
     */
    public ObjectWebOperatorGenerator(String wantedInterfacesPackage, String wantedModelsDirectory, ModelBindings modelBindings) {
        initialized = false;
        ignoredBindings = new LinkedList();
        collapsedElements = new LinkedList();

	this.interfacesPackage = wantedInterfacesPackage;
	this.modelsDirectory = wantedModelsDirectory;
        this.modelBindings = modelBindings;
    }

    /**
     * <p>
     *  This sets the output directory to output generated classes to.
     * </p>
     *
     * @param baseDir the name of the base directory where
     *        all generated classes should be placed
     */
    public void setOutputDirectory(String outputDir) {
        if (outputDir == null) {
            throw new IllegalArgumentException("A generator cannot have a " +
                "null output directory.");
        }

        this.outputDir = new File(outputDir);
        uninitialize();
    }

    public void setModelBindings(ModelBindings modelBindings) {
        if (modelBindings == null) {
            throw new IllegalArgumentException("A generator cannot have a null model hashtable");
        }
        this.modelBindings = modelBindings;
        uninitialize();
    }

    /**
     * <p>
     *  This sets the output directory to output generated classes to.
     * </p>
     *
     * @param outputDir the <code>File</code> (directory) where
     *        all generated classes should be placed
     */
    public void setOutputDirectory(File outputDir) {
        if (outputDir == null) {
            throw new IllegalArgumentException("A generator cannot have a " +
                "null output directory.");
        }

        this.outputDir = outputDir;
        uninitialize();
    }

    /**
     * <p>
     *  This allows a binding to be ignored. This is an intermediary step 
     *    towards allowing binding schemas. The name of the binding should 
     *    be passed in here, and no property generation for it will occur.
     * </p>
     *
     * @param xmlName the XML local name of the binding to ignore.
     * @param xmlNamespaceURI the XML namespace URI of the binding to ignore. 
     *        The empty string indicates there is no namespace associated 
     *        with the binding.
     */
    public void setIgnoreBinding(String xmlName, String xmlNamespaceURI) {
        ignoredBindings.add(new XMLName(xmlName, xmlNamespaceURI));
    }
    
    /**
     * <p>
     *  This allows a binding to be ignored. This is an intermediary step 
     *    towards allowing binding schemas. The name of the binding should 
     *    be passed in here, and no property generation for it will occur. This
     *    convenience version supplies an empty string for the namespace URI.
     * </p>
     *
     * @param xmlName the XML local name of the binding to ignore.
     */
    public void setIgnoreBinding(String xmlName) {
        setIgnoreBinding(xmlName, "");
    }
    
    /**
     * This will determine if a particular binding should be ignored. The
     * binding is identified by its XML local name and its XML namespace URI.
     *
     * @param xmlName the XML local name of the binding.
     * @param xmlNamespaceURI the XML namespace URI associated with the binding.
     *        An empty string indicates that there is no namespace associated 
     *        with the binding.
     *
     * @return true if and only if the binding should be ignored
     */
    protected boolean isIgnoredBinding(String xmlName, String xmlNamespaceURI) {
        if (ignoredBindings != null) {
            for (Iterator i = ignoredBindings.iterator(); i.hasNext(); ) {
                XMLName bindingName = (XMLName)i.next();
                if ((bindingName.getXMLName().equals(xmlName)) &&
                    (bindingName.getNamespaceURI().equals(xmlNamespaceURI))) {
                    return true;
                }
            }
        }

        return false;
    }

    /**
     * <p>
     *  This will convert from a set of constraints to Java classes.
     *    It accesses those constraints through the
     *    <code>{@link Binding}</code> interface, which provides
     *    them in a representation-independent format, and then
     *    converts them into Java code.
     * </p>
     *
     * @param binder <code>Binder</code> with ability to convert from
     *        constraints to Zeus <code>Binding</code>s.
     * @param root whether this is the root element
     * @throws <code>IOException</code> - when errors in writing to the
     *         supplied <code>Result</code> occur.
     * @throws <code>ZeusException</code> - when errors in class generation
     *         occur.
     */
    public void generate(Binding binding)
        throws IOException, ZeusException {

        collapsedElements.clear();

        if (!isInitialized()) {
            initialize();
        }

        /**
         * Currently, only Containers can be generated (which
         *   includes ContainerProperties), as AtomicProperties
         *   won't result in classes.
         */
        if (binding instanceof Container) {
            Container container = (Container)binding;

            // Ignore Java primitive types (from collapsed elements)
            if (ClassUtils.isJavaPrimitive(binding.getJavaType())) {
                return;
            }

            generateOperatorHandler(container);
            generateOperator(container);

        } else  {
            throw new UnsupportedBindingTypeException(binding);
        }
    }

    protected void generateOperatorHandler (Container container)
        throws IOException {

	List operatorHandlerImports = new LinkedList();

	// Get a new writer
	String filename = container.getJavaType() + "OperatorHandler.java";
	File outputFile = new File(getDir(container),filename);
	FileWriter writer = new FileWriter(outputFile);

	// Write out the source file header comment
	writer.write(ExtZeusDefaults.GENERATED_CLASS_COMMENT);

	// Add a package statement if necessary
	writePackageDeclaration(writer,
				container.getJavaInterfacePackage());

	operatorHandlerImports.add(this.interfacesPackage+".ExtensionImpl");
	operatorHandlerImports.add(this.interfacesPackage+".ExtensionManager");

	// Write out import statements
	writeImportStatements(writer, operatorHandlerImports);

	// Write out the implementation declaration
	writeDeclaration(writer, container,
		"class",
		"OperatorHandler",
		"ExtensionImpl",
		"");


	// Write out each property
	writer.write("    private ExtensionManager _extManager;\n\n");
	writer.write("    private ");
	writer.write(container.getJavaType());
	writer.write("Operator _operator;\n\n");

	// Write class constructor
	writer.write("    public ");
	writer.write(container.getJavaType());
	writer.write("OperatorHandler(");
	writer.write(") {\n");
	writer.write("        _operator = new ");
	writer.write(container.getJavaType());
	writer.write("Operator(this);\n");
	writer.write("    }\n");

	writer.write("    public ");
	writer.write("Object getInstance() {\n");
	writer.write("        return (_operator);\n");
	writer.write("    }\n\n");

	// Write out the methods involving properties
	writer.write("    public String toString() {\n");
	writer.write("        return(\"Operator\");\n");
	writer.write("    }\n\n");

	writer.write("}\n");

	writer.flush();
	writer.close();

    }

    /**
     * <p>
     *  This will generate the class containing informations
     *  about an instance in a operated state
     *  for the supplied <code>Container</code>.
     *  This generation is "cross-binding", using 2 DTDs;
     *  <code>Container</code> crossing in operator.dtd
     * </p>
     *
     * @param container <code>Container</code> to generate classes from.
     * @throws <code>IOException</code> - when generation errors occur.
     */
    protected void generateOperator(Container container)
        throws IOException {

        List properties = container.getProperties();

	// Get a new writer
	String filename = container.getJavaType() + "Operator.java";
	File outputFile = new File(getDir(container), filename);
	FileWriter writer = new FileWriter(outputFile);

	// Write out the source file header comment
	writer.write(ExtZeusDefaults.GENERATED_CLASS_COMMENT);

	// Add a package statement if necessary.
	writePackageDeclaration(writer,
				container.getJavaInterfacePackage());

        writeDeclaration(writer, container,
		"class",
		"Operator",
		"",
		"");

	// fields

	writer.write("    private ");
	writer.write(container.getJavaType());
	writer.write("OperatorHandler _handler;\n\n");

	//methods
        writer.write("    public ");
        writer.write(container.getJavaType());
        writer.write("Operator(");
	writer.write(container.getJavaType());
	writer.write("OperatorHandler handler) {\n");
	writer.write("        this._handler = handler ;\n");
	writer.write("    }\n\n");

        writer.write("    public void operate (String command, ");
        writer.write(this.interfacesPackage);
        writer.write(".Bean target) {\n");
        for (Iterator i= properties.iterator(); i.hasNext(); )
        {
            Property property = (Property)i.next();

          if ( (! isIgnoredBinding(property.getXMLName(),
                                 property.getXMLNamespaceURI())) &&
		(!property.getXMLName().startsWith("xmlns")) &&
		(!property.getModifier().get(Property.MUTABILITY_FINAL)) )
          {
            if (property.isCollection())
            {
                writer.write("        if (command.equals(\"remove");
                writer.write(CapitalizationUtils.initialUpper(property.getJavaName()));
                writer.write("\"))\n");
                writer.write("            ((");
                writer.write(container.getJavaType());
                writer.write("Bean)this.getManager().getSource()).remove");
                writer.write(CapitalizationUtils.initialUpper(property.getJavaName()));
                writer.write("( (");
                writer.write(CapitalizationUtils.initialUpper(property.getJavaType()));
                writer.write(") target);\n");
            }else if (property.getModifier().get(Property.SOURCE_ELEMENT)) {
                writer.write("        if (command.equals(\"set");
                writer.write(CapitalizationUtils.initialUpper(property.getJavaName()));
                writer.write("\"))\n");
                writer.write("            ((");
                writer.write(CapitalizationUtils.initialUpper(container.getJavaType()));
                writer.write("Bean)this.getManager().getSource()).set");
                writer.write(CapitalizationUtils.initialUpper(property.getJavaName()));
                writer.write("( new ");
                writer.write(CapitalizationUtils.initialUpper(property.getJavaType()));
                writer.write("ImplBean());\n");
            }
          }
        }
        for (Iterator i= properties.iterator(); i.hasNext(); )
        {
            Property property = (Property)i.next();
          if ( (! isIgnoredBinding(property.getXMLName(),
                                 property.getXMLNamespaceURI())) &&
		(!property.getXMLName().startsWith("xmlns")) &&
		(!property.getModifier().get(Property.MUTABILITY_FINAL)) )
          {
            if (property.isCollection())
            {
                writer.write("        if (command.equals(\"add");
                writer.write(CapitalizationUtils.initialUpper(property.getJavaName()));
                writer.write("\"))\n");
                writer.write("            ((");
                writer.write(container.getJavaType());
                writer.write("Bean)this.getManager().getSource()).add");
                writer.write(CapitalizationUtils.initialUpper(property.getJavaName()));
                writer.write("( new ");
                writer.write(CapitalizationUtils.initialUpper(property.getJavaType()));
                writer.write("ImplBean());\n");
            }else if (property.getModifier().get(Property.SOURCE_ELEMENT)) {
                writer.write("        if (command.equals(\"unset");
                writer.write(CapitalizationUtils.initialUpper(property.getJavaName()));
                writer.write("\"))\n");
                writer.write("            ((");
                writer.write(container.getJavaType());
                writer.write("Bean)this.getManager().getSource()).set");
                writer.write(CapitalizationUtils.initialUpper(property.getJavaName()));
                writer.write("(null);\n");
            }
          }
        }
	writer.write("        _handler.getManager().getSource().notifyListeners();\n");
        writer.write("    }\n\n");

	writer.write("    public ");
	writer.write(this.interfacesPackage);
	writer.write(".ExtensionManager getManager() {\n");
	writer.write("        return (_handler.getManager());\n");
	writer.write("    }\n\n");

        writer.write("}\n");

        writer.flush();
        writer.close();

    }

    /**
     * <p>
     *  This writes out the package declaration for this class, if needed.
     * </p>
     *
     * @param writer <code>Writer</code> to perform output to
     * @param packageName the package of this class
     * @throws <code>IOException</code> - when errors in output occur.
     */
    protected void writePackageDeclaration(Writer writer, String packageName)
        throws IOException {

        // Generate a package statement if necessary.
        if (!packageName.equals("")) {
            writer.write("package");
            writer.write(" ");
            writer.write(packageName);
            writer.write(";\n\n");
        }
    }

    /**
     * <p>
     *  This will take the supplied list of class names, and add import
     *    statements for each of the classes in the list.
     * </p>
     *
     * @param writer <code>Writer</code> to output to
     * @param importClasses <code>List</code> of classes to add imports for
     * @throws <code>IOException</code> - when errors in output occur.
     */
    protected void writeImportStatements(Writer writer,
                                                  List importClasses)
        throws IOException {

        // Write out imports
        if (importClasses.size() > 0) {
            writer.write("// Local Interface Import Statements\n");
        }
        for (Iterator i = importClasses.iterator(); i.hasNext(); ) {
            writer.write("import ");
            writer.write((String)i.next());
            writer.write(";\n");
        }
        if (importClasses.size() > 0) {
            writer.write("\n");
        }
    }
    
    /**
     * <p>
     *  This writes out the interface bean declaration (the class header).
     * </p>
     *
     * @param writer <code>Writer</code> to output to
     * @param container the container to write the interface for.
     * @throws <code>IOException</code> - when errors in output occur.
     */
    protected void writeDeclaration(Writer writer, 
                                             Container container,
					     String modifier,
					     String suffix,
					     String extended,
					     String implemented)
        throws IOException {
            
        writer.write("public ");
	writer.write(modifier);
	writer.write(" ");
        writer.write(container.getJavaType());
	writer.write(suffix);
	if (!extended.equals("")) {
	    writer.write(" extends ");
	    writer.write(extended);
	}
	
	if (!implemented.equals("")) {
	    writer.write(" implements ");
	    writer.write(implemented);
	}

	writer.write(" {\n\n");
    }
    
    /**
     * <p>
     *  This will initialize this instance of <code>Generator</code>,
     *    handling tasks like directory creation.
     * </p>
     *
     * @throws <code>IllegalOutputDirectoryException</code> - when an illegal
     *         output directory has been supplied to this <code>Generator</code>
     */    
    private void initialize() throws IllegalOutputDirectoryException {        
        // Determine the output directory
        if (outputDir == null) {
            outputDir = new File(".");
        }
        
        // Ensure the output directory exists and is a directory
        if ((outputDir.exists()) && (!outputDir.isDirectory())) {
            throw new IllegalOutputDirectoryException(
                outputDir.getAbsolutePath());
        }

        // Create the directory if it does not exist
        if (!outputDir.exists()) {
            outputDir.mkdir();
        }

        initialized = true;
    }        

    /**
     * <p>
     *  This will reset the initialization state of this <code>Generator</code>
     *    instance to <code>false</code>.
     * </p>
     */
    private void uninitialize() {
        initialized = false;
    }
    
    /**
     * <p>
     *  This will indicate if this <code>Generator</code> instance has been
     *    initialized.
     * </p>
     *
     * @return <code>boolean</code> - whether this instance is initialized.
     */
    private boolean isInitialized() {
        return initialized;
    }
    
    /**
     * <p>
     *  This will determine the directory to use for generation
     *    based on the packages of a <code>{@link Container}</code>.
     * </p>
     *
     * @param container the container to obtain the directory for.
     * @return <code>File</code> - the directory for generation.
     */
    private File getDir(Container container) {
        StringBuffer dirName = new StringBuffer();
        String thisPackage = container.getJavaInterfacePackage();

        for (int i=0, len=thisPackage.length(); i<len; i++) {
            char ch = thisPackage.charAt(i);
            if (ch != '.') {
                dirName.append(ch);
            } else {
                dirName.append(System.getProperty("file.separator"));
            }
        }
        
        File dir = 
            new File(outputDir, dirName.toString());
        
        // Create the directory if it does not exist
        if (!dir.exists()) {
            dir.mkdirs();
        }
        
        return dir;
    }
}
